@@ -97,10 +97,10 @@ char iso_esc_tab[][4] = {
  * Input
  */
 int in_ln_ascii(f, mi)
-register FILE *f;
+FILE *f;
 mlcinfo *mi;
 {
-    register int i;
+    int i;
 
     while (last < bufsize && ((i = getc(f)) != EOF) && i != '\n' && i != '\r')
 	buffer[last++] = i;
@@ -120,6 +120,8 @@ mlcinfo *mi;
 {
     if (Rctrl(c))
 	buffer[last++] = c;
+    else if (mi->GRset == MC_ASCII)
+	buffer[last++] = make_mlrep1(MC_ASCII,c);
     else
 	buffer[last++] = make_mlrep1(mi->GRset,c&0x7f);
 }
@@ -165,7 +167,12 @@ right_make(c, mi)
 textchar c;
 mlcinfo *mi;
 {
-    buffer[last++] = make_mlrep(mi->GRset,mi->inpend&0x7f,c&0x7f);
+    if (c&0x80) {
+	buffer[last++] = make_mlrep(mi->GRset,mi->inpend&0x7f,c&0x7f);
+    } else {
+	buffer[last++] = make_mlrep1(MC_ASCII,mi->inpend);
+	buffer[last++] = c;
+    }
     mi->instat = ISO_ONE;
 }
 
@@ -175,10 +182,15 @@ mlcinfo *mi;
 {
     textchar c1, c2;
 
-    c1 = mi->inpend&0x7f;
-    c2 = c&0x7f;
-    fix_kanji_range(&c1, &c2);
-    buffer[last++] = make_mlrep(mi->GRset,c1,c2);
+    if (c&0x80) {
+	c1 = mi->inpend&0x7f;
+	c2 = c&0x7f;
+	fix_kanji_range(&c1, &c2);
+	buffer[last++] = make_mlrep(mi->GRset,c1,c2);
+    } else {
+	buffer[last++] = make_mlrep1(MC_ASCII,mi->inpend);
+	buffer[last++] = c;
+    }
     mi->instat = ISO_ONE;
 }
 
@@ -287,8 +299,8 @@ mlcinfo *mi;
 	    mi->Gchsets[cr] = MC_BIG52;
 
 do_escape(f, mi)
-register FILE *f;
-register mlcinfo *mi;
+FILE *f;
+mlcinfo *mi;
 {
     int desreg;		/* designated register */
     int isonebyte;	/* one byte charset? */
@@ -358,8 +370,8 @@ register mlcinfo *mi;
 }
 
 do_GLlockshift(chreg, mi)
-register unsigned char chreg;
-register mlcinfo *mi;
+unsigned char chreg;
+mlcinfo *mi;
 {
     mi->GLinvk = chreg;
     mi->GLset = mi->Gchsets[chreg];
@@ -374,8 +386,8 @@ register mlcinfo *mi;
 }
 
 do_GRlockshift(chreg, mi)
-register unsigned char chreg;
-register mlcinfo *mi;
+unsigned char chreg;
+mlcinfo *mi;
 {
     mi->GRinvk = chreg;
     mi->GRset = mi->Gchsets[chreg];
@@ -405,10 +417,10 @@ register mlcinfo *mi;
 }
 
 int in_ln_iso(f, mi)
-register FILE *f;
-register mlcinfo *mi;
+FILE *f;
+mlcinfo *mi;
 {
-    register int i;
+    int i;
 
     while (last < bufsize &&
 	   ((i = getc(f)) != EOF) && i != '\n' && i != '\r') {
@@ -447,11 +459,16 @@ register mlcinfo *mi;
 /*
  * Output
  */
+#define	out_esc_seq(mc, f) \
+{ \
+    putc(ESC, f); fputs(iso_esc_tab[mc], f); \
+}
+
 Fputc(c, f)
 mlchar c;
 FILE *f;
 {
-    register mlcinfo *mi;
+    mlcinfo *mi;
 
     mi = &mlcodeinfo[fileno(f)];
     (*(mi->out_ch))(c, f, mi);
@@ -482,12 +499,18 @@ mlchar c;
 FILE *f;
 mlcinfo *mi;
 {
+    int mc;
     textchar c1, c2;
 
-    if (mlrep_mc(c) != mi->GRout)
+    if ((mc = mlrep_mc(c)) != mi->GRout)
 	return FALSE;
-    mlchar2euc(c, c1, c2);
-    putc(c1, f); putc(c2, f);
+    if (ml_font_blen(mc) == 2) {
+	mlchar2euc(c, c1, c2);
+	putc(c1, f); putc(c2, f);
+    } else {
+	mlchar2r(c, c1);
+	putc(c1, f);
+    }
     return TRUE;
 }
 
@@ -499,7 +522,7 @@ mlcinfo *mi;
     if (mlrep_mc(c) != MC_ASCII)
 	return FALSE;
     if (mi->GLset != MC_ASCII) {
-	putc(ESC, f); fputs(iso_esc_tab[MC_ASCII], f);
+	out_esc_seq(MC_ASCII, f);
 	mi->GLset = MC_ASCII;
     }
     putc(c, f);
@@ -517,7 +540,7 @@ mlcinfo *mi;
     mc = mlrep_mc(c);
     if (mc == MC_JP || mc == MC_JPOLD) {
 	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
-	    putc(ESC, f); fputs(iso_esc_tab[MC_ASCII], f);
+	    out_esc_seq(MC_ASCII, f);
 	    mi->GLset = MC_ASCII;
 	}
 	mlchar2sj(c, c1, c2);
@@ -525,7 +548,7 @@ mlcinfo *mi;
 	return TRUE;
     } else if (mc == MC_KANA) {
 	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
-	    putc(ESC, f); fputs(iso_esc_tab[MC_ASCII], f);
+	    out_esc_seq(MC_ASCII, f);
 	    mi->GLset = MC_ASCII;
 	}
 	putc(mlrep_char(c)|0x80, f);
@@ -545,7 +568,7 @@ mlcinfo *mi;
     mc = mlrep_mc(c);
     if (mc == MC_BIG51 || mc == MC_BIG52) {
 	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
-	    putc(ESC, f); fputs(iso_esc_tab[MC_ASCII], f);
+	    out_esc_seq(MC_ASCII, f);
 	    mi->GLset = MC_ASCII;
 	}
 	mlchar2bvh(mc, c, c1, c2);
@@ -566,7 +589,7 @@ mlcinfo *mi;
     mc = mlrep_mc(c);
     if (mc == MC_BIG51 || mc == MC_BIG52) {
 	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
-	    putc(ESC, f); fputs(iso_esc_tab[MC_ASCII], f);
+	    out_esc_seq(MC_ASCII, f);
 	    mi->GLset = MC_ASCII;
 	}
 	mlchar2bve(mc, c, c1, c2);
@@ -587,7 +610,7 @@ mlcinfo *mi;
     if ((*(mi->GRoutchar))(c, f, mi)) {
     } else {
 	if ((mc = mlrep_mc(c)) != mi->GLset) {
-	    putc(ESC, f); fputs(iso_esc_tab[mc], f);
+	    out_esc_seq(mc, f);
 	}
 	if (ml_font_blen(mc) == 2) {
 	    mlchar2iso(c, c1, c2);
@@ -599,74 +622,326 @@ mlcinfo *mi;
     }
 }
 
+/*
+ * Dvi-special
+ */
+/*
+#define	specchstart	spec_ch_start
+#define	specch		spec_ch
+#define	specpoollen	spec_pool_len
+#define	getspecch	get_spec_ch
+*/
+#define	spec_ch_start	specchstart
+#define	spec_ch		specch
+#define	spec_pool_len	specpoollen
+#define	get_spec_ch	getspecch
+
+#define	SPEC_POOL_SIZE	2048
+mlcinfo spec_mlcodeinfo;
+textchar *spec_pool = NULL;
+textchar *spec_ptr = NULL;
+textchar *spec_end = NULL;
+int spec_pool_size;
+
+#define	spec_esc_seq(mc) \
+{ \
+    textchar *e; \
+    specc(ESC); \
+    for (e = iso_esc_tab[mc]; *e != '\0'; e++) \
+	specc(*e); \
+}
+
+specc(c)
+mlchar c;
+{
+    textchar *p;
+    int size;
+
+    if (spec_ptr >= spec_end) {
+	size = spec_pool_size+SPEC_POOL_SIZE;
+	if ((p = realloc(spec_pool, size)) == NULL) {
+	    fprintf(stderr, "no spec pool\n");
+	    exit(1);
+	}
+	spec_pool = p;
+	spec_pool_size = size;
+	spec_end = spec_pool+spec_pool_size;
+    }
+    *spec_ptr++ = c;
+}
+
+spec_ch_start()
+{
+    if (spec_pool == NULL) {
+	spec_pool_size = SPEC_POOL_SIZE;
+	spec_pool = malloc(spec_pool_size);
+	spec_end = spec_pool+spec_pool_size;
+    }
+    spec_ptr = spec_pool;
+}
+
+spec_ch(c)
+mlchar c;
+{
+    mlcinfo *mi;
+
+    mi = &spec_mlcodeinfo;
+    (*(mi->out_ch))(c, mi);
+}
+
+spec_ch_ascii(c, mi)
+mlchar c;
+mlcinfo *mi;
+{
+    if (is_mlrep_ascii(c)) {
+	specc(c);
+    } else { /* TODO */
+	/*fprintf(f, "[%x|%x]", mlrep_mc(c), mlrep_char(c));*/
+    }
+}
+
+no_spec_ch(c, mi)
+mlchar c;
+mlcinfo *mi;
+{
+    return FALSE;
+}
+
+right_spec_ch(c, mi)
+mlchar c;
+mlcinfo *mi;
+{
+    int mc;
+    textchar c1, c2;
+
+    if ((mc = mlrep_mc(c)) != mi->GRout)
+	return FALSE;
+    if (ml_font_blen(mc) == 2) {
+	mlchar2euc(c, c1, c2);
+	specc(c1); specc(c2);
+    } else {
+	mlchar2r(c, c1);
+	specc(c1);
+    }
+    return TRUE;
+}
+
+eight_spec_ch(c, mi)
+mlchar c;
+mlcinfo *mi;
+{
+    if (mlrep_mc(c) != MC_ASCII)
+	return FALSE;
+    if (mi->GLset != MC_ASCII) {
+	spec_esc_seq(MC_ASCII);
+	mi->GLset = MC_ASCII;
+    }
+    specc(c);
+    return TRUE;
+}
+
+sjis_spec_ch(c, mi)
+mlchar c;
+mlcinfo *mi;
+{
+    int mc;
+    textchar c1, c2;
+
+    mc = mlrep_mc(c);
+    if (mc == MC_JP || mc == MC_JPOLD) {
+	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
+	    spec_esc_seq(MC_ASCII);
+	    mi->GLset = MC_ASCII;
+	}
+	mlchar2sj(c, c1, c2);
+	specc(c1); specc(c2);
+	return TRUE;
+    } else if (mc == MC_KANA) {
+	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
+	    spec_esc_seq(MC_ASCII);
+	    mi->GLset = MC_ASCII;
+	}
+	specc(mlrep_char(c)|0x80);
+	return TRUE;
+    } else
+	return FALSE;
+}
+
+big5h_spec_ch(c,mi)
+mlchar c;
+mlcinfo *mi;
+{
+    int mc;
+    textchar c1, c2;
+
+    mc = mlrep_mc(c);
+    if (mc == MC_BIG51 || mc == MC_BIG52) {
+	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
+	    spec_esc_seq(MC_ASCII);
+	    mi->GLset = MC_ASCII;
+	}
+	mlchar2bvh(mc, c, c1, c2);
+	specc(c1); specc(c2);
+	return TRUE;
+    } else
+	return FALSE;
+}
+
+big5e_spec_ch(c, f, mi)
+mlchar c;
+FILE *f;
+mlcinfo *mi;
+{
+    int mc;
+    textchar c1, c2;
+
+    mc = mlrep_mc(c);
+    if (mc == MC_BIG51 || mc == MC_BIG52) {
+	if (mi->GLset != mc && mi->GLset != MC_ASCII) {
+	    spec_esc_seq(MC_ASCII);
+	    mi->GLset = MC_ASCII;
+	}
+	mlchar2bve(mc, c, c1, c2);
+	specc(c1); specc(c2);
+	return TRUE;
+    } else
+	return FALSE;
+}
+
+spec_ch_iso(c, mi)
+mlchar c;
+mlcinfo *mi;
+{
+    int mc;
+    textchar *e, c1, c2;
+
+    if ((*(mi->GRoutchar))(c, mi)) {
+    } else {
+	if ((mc = mlrep_mc(c)) != mi->GLset) {
+	    spec_esc_seq(mc);
+	}
+	if (ml_font_blen(mc) == 2) {
+	    mlchar2iso(c, c1, c2);
+	    specc(c1); specc(c2);
+	} else {
+	    specc(mlrep_char(c));
+	}
+	mi->GLset = mc;
+    }
+}
+
+spec_pool_len()
+{
+    return (spec_ptr-spec_pool);
+}
+
+textchar
+get_spec_ch()
+{
+    return *spec_ptr++;
+}
+
 /*
  * Initialize
  */
-int hasargcode = false;
-static int argcode;
-
-argmlcode(code)
-char *code;
+defaultiomlcodetype(in, out)
+FILE *in, *out;
 {
-    int iso, g0, g1, c, d, iscode;
+    setinmlcodetype(in, make_ict(0,MC_ASCII,MC_ASCII));
+    setoutmlcodetype(out, make_oct(0,MC_ASCII));
+}
+
+initallmlcodetype()
+{
+    setallmlcodetype(make_ict(1,MC_ASCII,MC_ASCII), make_oct(1,MC_ASCII));
+}
+
+str2ct(code, mkct, isct)
+char *code;
+int (*mkct)();
+int *isct;
+{
+    int c, d;
     char *s;
 
-    hasargcode = true;
+    *isct = true;
     if (strcmp(code, "ascii") == 0 || strcmp(code, "tex") == 0)
-	argcode = make_ict(0,MC_ASCII,MC_ASCII);
+	return (*mkct)(0,MC_ASCII);
+    else if (strcmp(code, "jis7") == 0 || strcmp(code, "iso-2022-7bit") == 0)
+	return (*mkct)(1,MC_ASCII);
     else if (strcmp(code, "junet") == 0 || strcmp(code, "euc") == 0 ||
-	     strcmp(code, "euc-japan") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_JP);
-    else if (strcmp(code, "euc-china") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_CN);
+	     strcmp(code, "euc-japan") == 0 || strcmp(code, "euc-jp") == 0)
+	return (*mkct)(1,MC_JP);
+    else if (strcmp(code, "euc-china") == 0 || strcmp(code, "euc-cn") == 0)
+	return (*mkct)(1,MC_CN);
     else if (strcmp(code, "euc-korea") == 0 || strcmp(code, "euc-kr") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_KR);
+	return (*mkct)(1,MC_KR);
     else if (strncmp(code, "latin", 5) == 0) {
 	c = *(code+5)-'0';
 	if (1 <= c && c <= 4)
-	    argcode = make_ict(1,MC_ASCII,MC_LTN1+c-1);
+	    return (*mkct)(1,MC_LTN1+c-1);
 	else if (c == 5)
-	    argcode = make_ict(1,MC_ASCII,MC_LTN5);
+	    return (*mkct)(1,MC_LTN5);
 	else
-	    argcode = make_ict(0,MC_ASCII,MC_LTN1);
+	    return (*mkct)(0,MC_LTN1);
     } else if (strncmp(code, "iso-8859-", 9) == 0) {
 	c = *(code+9)-'0';
 	if (1 <= c && c <= 4)
-	    argcode = make_ict(1,MC_ASCII,MC_LTN1+c-1);
+	    return (*mkct)(1,MC_LTN1+c-1);
 	else if (c == 5)
-	    argcode = make_ict(1,MC_ASCII,MC_LTN5);
+	    return (*mkct)(1,MC_LTN5);
 	else
-	    argcode = make_ict(1,MC_ASCII,MC_LTN1);
+	    return (*mkct)(1,MC_LTN1);
     } else if (strcmp(code, "grk") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_GRK);
+	return (*mkct)(1,MC_GRK);
     else if (strcmp(code, "crl") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_CRL);
+	return (*mkct)(1,MC_CRL);
     else if (strcmp(code, "sjis") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_SJIS);
+	return (*mkct)(1,MC_SJIS);
     else if (strcmp(code, "bigv") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_BIG5h);
+	return (*mkct)(1,MC_BIG5h);
     else if (strcmp(code, "bigvet") == 0)
-	argcode = make_ict(1,MC_ASCII,MC_BIG5e);
+	return (*mkct)(1,MC_BIG5e);
     else {
-	for (iscode = true, c = 0, s = code; *s != '\0'; s++) {
+	for (*isct = true, c = 0, s = code; *s != '\0'; s++) {
 	    if ('0' <= *s && *s <= '9')
 		d = *s-'0';
 	    else if ('a' <= *s && *s <= 'f')
 		d = *s-'a'+10;
+	    else if ('A' <= *s && *s <= 'F')
+		d = *s-'A'+10;
 	    else {
-		iscode = false;
+		*isct = false;
 		break;
 	    }
 	    c = c*16+d;
 	}
-	if (iscode)
-	    argcode = c;
-	else {
-	    argcode = make_ict(1,MC_ASCII,MC_JP);
-	    fprintf(stderr,
-		    "No such file type %s, I will change it to junet.\n",
-		    code);
-	}
+	return c;
+    }
+}
+
+int hasargcode = false;
+static int argcode;
+
+make_inct(iso,gr)
+int iso, gr;
+{
+    return (make_ict(iso,MC_ASCII,gr));
+}
+
+argmlcode(code)
+char *code;
+{
+    int isct;
+
+    hasargcode = true;
+    argcode = str2ct(code, make_inct, &isct);
+    /* TODO: range check of argcode */
+    if (!isct) {
+	argcode = make_inct(1,MC_JP);
+	fprintf(stderr,
+		"No such file type %s, I will change it to junet.\n",
+		code);
     }
 }
 
@@ -675,11 +950,65 @@ getargcode()
     return argcode;
 }
 
+int hasargterm = false;
+static int argterm;
+
+make_outct(iso,gr)
+int iso, gr;
+{
+    return (make_oct(iso,gr));
+}
+
+argmlterm(code)
+char *code;
+{
+    int isct;
+
+    hasargterm = true;
+    argterm = str2ct(code, make_outct, &isct);
+    /* TODO: range check of argterm */
+    if (!isct) {
+	argterm = make_outct(1,MC_ASCII);
+	fprintf(stderr,
+		"No such terminal type %s, I will change it to iso-2022-7bit.\n",
+		code);
+    }
+}
+
+getargterm()
+{
+    return argterm;
+}
+
+int hasarglog = false;
+static int arglog;
+
+argmllog(code)
+char *code;
+{
+    int isct;
+
+    hasarglog = true;
+    arglog = str2ct(code, make_outct, &isct);
+    /* TODO: range check of arglog */
+    if (!isct) {
+	arglog = make_outct(1,MC_ASCII);
+	fprintf(stderr,
+		"No such log type %s, I will change it to iso-2022-7bit.\n",
+		code);
+    }
+}
+
+getarglog()
+{
+    return arglog;
+}
+
 setinmlcodetype(f, mtype)
 FILE *f;
 int mtype;
 {
-    register mlcinfo *mi = &mlcodeinfo[fileno(f)];
+    mlcinfo *mi = &mlcodeinfo[fileno(f)];
 
     mi->iso = ict_iso(mtype);
     mi->G0set = ict_G0(mtype);
@@ -701,11 +1030,11 @@ setoutmlcodetype(f, mtype)
 FILE *f;
 int mtype;
 {
-    register mlcinfo *mi = &mlcodeinfo[fileno(f)];
+    mlcinfo *mi = &mlcodeinfo[fileno(f)];
     int gr;
 
     mi->GRout = gr = oct_GR(mtype);
-    if (gr == MC_NONE)
+    if (gr == MC_NONE || gr == MC_ASCII)
 	mi->GRoutchar = no_out_ch;
     else if (gr == MC_EIGHT)
 	mi->GRoutchar = eight_out_ch;
@@ -723,6 +1052,55 @@ int mtype;
 	mi->out_ch = out_ch_ascii;
 }
 
+int hasargspec = false;
+static int argspec;
+
+argmlspec(code)
+char *code;
+{
+    int isct;
+
+    hasargspec = true;
+    argspec = str2ct(code, make_outct, &isct);
+    /* TODO: range check of argspec */
+    if (!isct) {
+	argspec = make_outct(1,MC_JP);
+	fprintf(stderr,
+		"No such special type %s, I will change it to euc-jp.\n",
+		code);
+    }
+}
+
+getargspec()
+{
+    return argspec;
+}
+
+setspecmlcodetype(mtype)
+int mtype;
+{
+    mlcinfo *mi = &spec_mlcodeinfo;
+    int gr;
+
+    mi->GRout = gr = oct_GR(mtype);
+    if (gr == MC_NONE || gr == MC_ASCII)
+	mi->GRoutchar = no_spec_ch;
+    else if (gr == MC_EIGHT)
+	mi->GRoutchar = eight_spec_ch;
+    else if (gr == MC_SJIS)
+	mi->GRoutchar = sjis_spec_ch;
+    else if (gr == MC_BIG5h)
+	mi->GRoutchar = big5h_spec_ch;
+    else if (gr == MC_BIG5e)
+	mi->GRoutchar = big5e_spec_ch;
+    else
+	mi->GRoutchar = right_spec_ch;
+    if (oct_iso(mtype))
+	mi->out_ch = spec_ch_iso;
+    else
+	mi->out_ch = spec_ch_ascii;
+}
+
 setinmlmcs(t, fin, mcs)
 int t, fin, mcs;
 {
